/*! \file DynWardrobe.cpp

    \brief DynRPG plugin to combine multiple Animation2 actor images at run-time based on actor equipment

    \author Bernard J. Badger, AKA Aubrey the Bard

    DynWardrobe is a DynRPG plugin which combines multiple Animation2 images in layers at run-time
    to create new images for actors (aka heroes) in battle. The images are selected based on the
    items which an actor has equipped at the time, thus allowing equipment to have a visible
    presence in battle.
*/

//#define DYNWARDROBE_DEBUG              // Comment out to remove debugging

#include <DynRPG/DynRPG.h>
#include <cstdlib>
#include <set>
#include <string>
#include <fstream>
#include "CSVIterator.h"
#ifdef DYNWARDROBE_DEBUG
#include <sstream>          // Only needed for debug purposes
#include <iostream>         // Only needed for debug purposes
#endif // DYNWARDROBE_DEBUG

// Use this as a template for debug popups
// DEBUG
#ifdef DYNWARDROBE_DEBUG
//std::stringstream message;
//message << "message";
//MessageBox(
//            NULL, // We don't need a window handle
//            message.str().c_str(), // Text
//            "Debug", // Title
//            MB_OK // Flags
//);
#endif // DYNWARDROBE_DEBUG
// /DEBUG

// Use this as a template for writing to the debug file
// DEBUG
#ifdef DYNWARDROBE_DEBUG
//*ofDebug << "message\n";
#endif // DYNWARDROBE_DEBUG
// /DEBUG

//! Equipment slot index
/*!
    This enumerator provides indexing for the different equipment slots in which an item can be
    found on an RM2K3 battler.
*/
enum EquipSlot
{
    INVALID_SLOT,                                       //!< Invalid
    ACCESSORY_SLOT,                                     //!< Accessory
    ARMOR_SLOT,                                         //!< Armor
    HELMET_SLOT,                                        //!< Helmet
    WEAPON_SLOT,                                        //!< Weapon
    SHIELD_SLOT                                         //!< Shield/second weapon
};

//! Associator for a wardrobe article
/*!
    This class records the associations necessary for a wardrobe article. The associations consist of
    these components:

    The layer on which the item's Animation2 images should be drawn
    The RM2K3 equipment slot (weapon, shield, armor, helmet, or accessory)
    The Animation2 to add to the composite Animation2 (the item)
    The group of RM2K3 Animation2's to which the base Animation2 (the actor's) belongs
    The group of RM2K3 items to which the equipped item belongs
*/
class WardrobeAssociator
{

public:

    //! Default constructor
    /*!
        The default constructor of WardrobeAssociator provides an empty instance.

        \return (WardrobeAssociator &) Reference to this instance
    */
    WardrobeAssociator()
    {
        // Initialize variables
        mDrawingLayer = 0;
        mEquipSlot = INVALID_SLOT;
        mItemAnim2Index = -1;
        mActorAnim2Indices = std::set<int>();
        mItemIndices = std::set<int>();
    }

    //! Copy constructor
    /*!
        The copy constructor of WardrobeAssociator copies a given WardrobeAssociator.

        \param copyin : (const WardrobeAssociator &) Reference to object to be copied.
        \return (WardrobeAssociator &) Reference to this instance
    */
    WardrobeAssociator( const WardrobeAssociator & copyin )
    {
        mDrawingLayer = copyin.mDrawingLayer;
        mEquipSlot = copyin.mEquipSlot;
        mItemAnim2Index = copyin.mItemAnim2Index;
        mActorAnim2Indices = copyin.mActorAnim2Indices;
        mItemIndices = copyin.mItemIndices;
    }

    //! Destructor
    /*!
        The destructor of WardrobeAssociator deletes all member objects and the WardrobeAssociator
        object itself.
    */
    ~WardrobeAssociator()
    {
    }

    //! Overloaded = operator
    /*!
        The = operator for this class turns this instance into a deep copy of the right-hand side.

        \param rhs : (const WardrobeAssociator &) Reference to right-hand side
        \return (WardrobeAssociator &) Reference to this instance
    */
    WardrobeAssociator& operator=( const WardrobeAssociator & rhs )
    {
        mDrawingLayer = rhs.mDrawingLayer;
        mEquipSlot = rhs.mEquipSlot;
        mItemAnim2Index = rhs.mItemAnim2Index;
        mActorAnim2Indices = rhs.mActorAnim2Indices;
        mItemIndices = rhs.mItemIndices;
        return *this;
    }

    //! Compare index sets
    /*!
        This method takes two sets of integers (like the sets of indices for actor Animation2s) and
        compares them, considering the lesser set to be the one which contains the lowest value not
        matched by the other set, or if one set is a proper subset of the other, the subset is
        considered lesser. The method returns -1 iff rSet1 is lesser, 0 iff rSet1 is equivalent to
        rSet2, and 1 iff rSet2 is lesser. This method is used by the overloaded == operator and <
        operator of this class.

        \param rSet1 : (const std::set<int> &) Reference to first set for comparison
        \param rSet2 : (const std::set<int> &) Reference to second set for comparison
        \return (int) -1 iff rSet1 is lesser, 0 iff rSet1 is equivalent to rSet2, 1 iff rSet2 is lesser
    */
    static int IndexSetComparison( const std::set<int> &rSet1, const std::set<int> &rSet2 )
    {
        static std::set<int>::iterator iter1, iter2;       // Iterators for traversing sets

        // Compare sets
        iter1 = rSet1.begin();
        iter2 = rSet2.begin();
        for( ; ; iter1 != rSet1.end() && iter2 != rSet2.end() )
        {
            if( iter1 == rSet1.end() )
            {   // Set 1 ended before set 2; return 'set 1 is lesser' signal
                return -1;
            }
            if( iter1 == rSet2.end() )
            {   // Set 2 ended before set 1; return 'set 2 is lesser' signal
                return 1;
            }
            if( *iter1 < *iter2 )
            {   // Element of set 1 less than element of set 2 found; return 'set 1 is lesser' signal
                return -1;
            }
            if( *iter1 > *iter2 )
            {   // Element of set 2 less than element of set 1 found; return 'set 2 is lesser' signal
                return 1;
            }
            ++iter1;
            ++iter2;
        }
        // Comparison failed to detect any difference between sets; return 'equivalent' signal
        return 0;
    }

    //! Overloaded == operator
    /*!
        The == operator for this class returns 1 iff all member variables for this instance are
        equal to all member variables for the right-hand side.

        \param rhs : (const WardrobeAssociator &) Reference to right-hand side
        \return (int) 0 for false, 1 for true
    */
    int operator==( const WardrobeAssociator &rhs ) const
    {
        if( mDrawingLayer != rhs.mDrawingLayer
            || mEquipSlot != rhs.mEquipSlot
            || mItemAnim2Index != rhs.mItemAnim2Index
            || IndexSetComparison( mActorAnim2Indices, rhs.mActorAnim2Indices )
            || IndexSetComparison( mItemIndices, rhs.mItemIndices ) )
        {   // Member variables are not equivalent; return false
            return 0;
        }
        // Member variables are equivalent; return true
        return 1;
    }

    //! Overloaded < operator
    /*!
        The < operator for this class returns 1 iff this WardrobeAssociator is considered less than
        the right-hand side. It is important that the WardrobeAssociators are compared according to
        the order of priority of the data which determines which Animation2s are drawn first when
        creating new actor images, so that the set of WardrobeAssociators is sorted properly. The
        order of priority is as follows:

        mDrawingLayer: The user-defined drawing layer, giving first priority to user instruction
        mEquipSlot: The equipment slot, sorted as defined in the enum EquipSlot
        mItemAnim2Index: The index of the Animation2 being drawn in, of arbitrary importance
        mActorAnim2Indices: The indices of actor-associated Animation2s, of arbitrary importance
        mItemIndices: The indices of items, of arbitrary importance

        \param rhs : (const WardrobeAssociator &) Reference to right-hand side
        \return (int) 0 for false, 1 for true
    */
    int operator<( const WardrobeAssociator &rhs ) const
    {
        // First compare mDrawingLayer
        if( mDrawingLayer < rhs.mDrawingLayer )
        {   // This object's mDrawingLayer is lesser than the other object's mDrawingLayer
            return 1;
        }
        if( rhs.mDrawingLayer < mDrawingLayer )
        {   // The other object's mDrawingLayer is lesser than this object's mDrawingLayer
            return 0;
        }
        // mDrawingLater equivalent; next compare mEquipSlot
        if( mEquipSlot < rhs.mEquipSlot )
        {   // This object's mEquipSlot is lesser than the other object's mEquipSlot
            return 1;
        }
        if( rhs.mEquipSlot < mEquipSlot )
        {   // The other object's mEquipSlot is lesser than this object's mEquipSlot
            return 0;
        }
        // mEquipSlot equivalent; next compare mItemAnim2Index
        if( mItemAnim2Index < rhs.mItemAnim2Index )
        {   // This object's mItemAnim2Index is lesser than the other object's mItemAnim2Index
            return 1;
        }
        if( rhs.mItemAnim2Index < mItemAnim2Index )
        {   // The other object's mItemAnim2Index is lesser than this object's mItemAnim2Index
            return 0;
        }
        // mItemAnim2Index equivalent; next compare mActorAnim2Indices
        if( -1 == IndexSetComparison( mActorAnim2Indices, rhs.mActorAnim2Indices ) )
        {   // This object's mActorAnim2Indices is lesser than the other object's mActorAnim2Indices
            return 1;
        }
        if( -1 == IndexSetComparison( rhs.mActorAnim2Indices, mActorAnim2Indices ) )
        {   // The other object's mActorAnim2Indices is lesser than this object's mActorAnim2Indices
            return 0;
        }
        // mActorAnim2Indices equivalent; next compare mItemIndices
        if( -1 == IndexSetComparison( mItemIndices, rhs.mItemIndices ) )
        {   // This object's mItemIndices is lesser than the other object's mItemIndices
            return 1;
        }
        if( -1 == IndexSetComparison( rhs.mItemIndices, mItemIndices ) )
        {   // The other object's mItemIndices is lesser than this object's mItemIndices
            return 0;
        }
        // No differences found
        return 0;
    }

    //! Clear the data of this WardrobeAssociator
    /*!
        clear resets the data values of this WardrobeAssociator to defaults.
    */
    void clear()
    {
        mDrawingLayer = 0;
        mEquipSlot = INVALID_SLOT;
        mItemAnim2Index = -1;
        mActorAnim2Indices.clear();
        mItemIndices.clear();
    }

    //! Access mDrawingLayer
    /*!
        Access the current value of mDrawingLayer.

        \return (int) The current value of mDrawingLayer
    */
    int GetDrawingLayer() const
    {
        return mDrawingLayer;
    }

    //! Set mDrawingLayer
    /*!
        Set mDrawingLayer to the given value.

        \param rValue: (int) New value to set
    */
    void SetDrawingLayer( int rValue )
    {
        mDrawingLayer = rValue;
    }

    //! Access mEquipSlot
    /*!
        Access the current value of mEquipSlot.

        \return (EquipSlot) The current value of mEquipSlot
    */
    EquipSlot GetEquipSlot() const
    {
        return mEquipSlot;
    }

    //! Set mEquipSlot
    /*!
        Set mEquipSlot to the given value.

        \param rValue: (EquipSlot) New value to set
    */
    void SetEquipSlot( EquipSlot rValue )
    {
        mEquipSlot = rValue;
    }

    //! Access mItemAnim2Index
    /*!
        Access the current value of mItemAnim2Index.

        \return (int) The current value of mItemAnim2Index
    */
    int GetItemAnim2Index() const
    {
        return mItemAnim2Index;
    }

    //! Set mItemAnim2Index
    /*!
        Set mItemAnim2Index to the given value.

        \param rValue: (int) New value to set
    */
    void SetItemAnim2Index( int rValue )
    {
        mItemAnim2Index = rValue;
    }

    //! Set mActorAnim2Indices
    /*!
        Set mActorAnim2Indices to a copy of the given set.

        \param rValue: (const std::set<int> &) New value to set
    */
    void SetActorAnim2Indices( const std::set<int> &rValue )
    {
        mActorAnim2Indices = rValue;
    }

    //! Count the number of elements in mActorAnim2Indices with a given value
    /*!
        CountActorAnim2Indices returns the result of calling std::set::count on mActorAnim2Indices.

        \param rValue: (int) The value to check for
        \return (int) The number of elements in mActorAnim2Indices with the given value
    */
    int CountActorAnim2Indices( int rValue ) const
    {
        return mActorAnim2Indices.count( rValue );
    }

    //! Set mItemIndices
    /*!
        Set mItemIndices to a copy of the given set.

        \param rValue: (const std::set<int> &) New value to set
    */
    void SetItemIndices( const std::set<int> &rValue )
    {
        mItemIndices = rValue;
    }

    //! Count the number of elements in mItemIndices with a given value
    /*!
        CountActorAnim2Indices returns the result of calling std::set::count on mItemIndices.

        \param rValue: (int) The value to check for
        \return (int) The number of elements in mItemIndices with the given value
    */
    int CountItemIndices( int rValue ) const
    {
        return mItemIndices.count( rValue );
    }

    //! Validate the data in this WardrobeAssociator
    /*!
        IsValid returns true iff the data in this WardrobeAssociator is complete (no empty sets)
        and valid (no out-of-bound indices or invalid equipment slots). This is important even if
        there is no user error in the data file from which the data is drawn, because there could be
        extra "empty" lines in the DynWardrobe.csv file generated because of the "Equipment Slot
        Names" data off to the side in DynWardrobe.xsl.

        \return (bool) true if the data in this WardrobeAssociator is valid, false otherwise
    */
    bool IsValid()
    {
        static std::set<int>::iterator iter;        // Iterator for traversing sets

        if( INVALID_SLOT == mEquipSlot )
            return false;
        if( 0 > mItemAnim2Index || RPG::battleAnimations.count() < mItemAnim2Index )
            return false;
        if( mActorAnim2Indices.empty() )
            return false;
        for( iter = mActorAnim2Indices.begin(); iter != mActorAnim2Indices.end(); ++iter )
        {
            if( 0 > *iter || RPG::battleAnimations.count() < *iter )
                return false;
        }
        if( mItemIndices.empty() )
            return false;
        for( iter = mItemIndices.begin(); iter != mItemIndices.end(); ++iter )
        {
            if( 0 > *iter || RPG::items.count() < *iter )
                return false;
        }
        // All validation checks passed; return true
        return true;
    }

private:

    int mDrawingLayer;                                  //!< The layer on which the item's Animation2 images should be drawn
    EquipSlot mEquipSlot;                               //!< RM2K3 equipment slot (weapon, shield, armor, helmet, or accessory)
    int mItemAnim2Index;                                //!< Index of RM2K3 Animation2 to add to the composite Animation2 (the item)
    std::set<int> mActorAnim2Indices;                   //!< Set of indices of base (actor's) RM2K3 Animation2s which are part of this association
    std::set<int> mItemIndices;                         //!< Set of indices of RM2K3 items which are part of this association
};

//! Indices of data in the .csv file for wardrobe associations
enum WardrobeCSVIndex
{
    ITEM_CSV_INDEX = 1,                                 //!< Zero-based index of a wardrobe association's "item index" data in the .csv file
    EQUIP_SLOT_CSV_INDEX = 2,                           //!< Zero-based index of a wardrobe association's "equipment slot" data in the .csv file
    ITEM_ANIM2_CSV_INDEX = 4,                           //!< Zero-based index of a wardrobe association's "item Animation2 index" data in the .csv file
    DRAWING_LAYER_CSV_INDEX = 5,                        //!< Zero-based index of a wardrobe association's "drawing layer" data in the .csv file
    ACTOR_ANIM2_CSV_INDEX = 7                           //!< Zero-based index of a wardrobe association's "actor Animation2 index" data in the .csv file
};

const int POSE_WIDTH = 144;                             //!< Default value of poseWidth
const int POSE_HEIGHT = 48;                             //!< Default value of poseHeight

const int NUM_HEROES = 4;                               //!< Maximum number of heroes
const int NUM_ACTOR_POSES = 32;                         //!< Number of pose images for an actor

int poseWidth = POSE_WIDTH;                             //!< Width of an Actor's pose image
int poseHeight = POSE_HEIGHT;                           //!< Height of an Actor's pose image

std::set<WardrobeAssociator> wardrobe;                  //!< Set of WardrobeAssociators

bool inBattle;                                          //!< Whether the game is currently in a battle

#ifdef DYNWARDROBE_DEBUG
std::ofstream * ofDebug;                                //!< File output stream to record debug data
#endif // DYNWARDROBE_DEBUG

//! Draw an Animation2 onto the battle poses of an actor
/*!
    drawAnimation2 draws the Animation2 indicated by a given index onto the battle pose images of a
    given actor.

    \param rAnim2Index: (int) Index of the Animation2 to draw
    \param rActorPtr: (RPG::Actor *) Pointer to the actor to draw upon
*/
void drawAnim2( int rAnim2Index, RPG::Actor * rActorPtr )
{
    static int poseIndex;                   // Index of the current pose
    static int sourceX, sourceY;            // Coordinates within source image file
    static RPG::Image *sourceImagePtr;      // Source Image for drawing
    static std::string prevFilename = "";   // Previous file name for source image
    static std::string curFilename;         // Current file name for source image
    static std::string fullFilename;        // Full file name including folder and extension
    static std::ifstream testStream;        // Input file stream for testing whether a file exists

    // Initialize the source Image
    sourceImagePtr = RPG::Image::create();
    // Loop through all pose images of the actor
    for( poseIndex = 0; poseIndex < NUM_ACTOR_POSES; poseIndex++ )
    {
        // Make sure the Animation2's pose animation type is character pose; do nothing otherwise
        if( RPG::PT_CHARACTER_POSE == RPG::battlerAnimations[rAnim2Index]->poses[poseIndex]->animType )
        {   // Pose animation type is character pose
            // Get the name of the Animation2's image file
            curFilename = RPG::battlerAnimations[rAnim2Index]->poses[poseIndex]->battleCharsetFilename.s_str();
            // Check for non-existent image name
            if( 0 != curFilename.compare( "" ) )
            {   // Image name exists
                // Check whether the source image file has changed since last iteration
                if( 0 != curFilename.compare( prevFilename ) )
                {   // Source image file changed
                    // Add path and file type extension to file name
                    // Check for file as a png
                    fullFilename = "BattleCharSet\\" + curFilename + ".png";
                    testStream.open( fullFilename.c_str() );
                    if( testStream )
                    {   // File exists as png
                        testStream.close();
                    }
                    else
                    {   // File does not exist as png
                        // Check for file as a bmp
                        fullFilename = "BattleCharSet\\" + curFilename + ".bmp";
                        testStream.open( fullFilename.c_str() );
                        if( testStream )
                        {   // File exists as bmp
                            testStream.close();
                        }
                        else
                        {   // File does not exist as bmp
                            // Check for file as an xyz
                            fullFilename = "BattleCharSet\\" + curFilename + ".xyz";
                            testStream.open( fullFilename.c_str() );
                            if( testStream )
                            {   // File exists as xyz
                                testStream.close();
                            }
                            else
                            {   // File does not exist as xyz
                                fullFilename = "";
                            }
                        }
                    }
                    if( 0 == fullFilename.compare( "" ) )
                    {   // File was not found
                        sourceImagePtr->clear();
                    }
                    else
                    {   // File was found
                        // Load source Image from file
                        sourceImagePtr->loadFromFile( fullFilename );
                    }
                    // Remember file name
                    prevFilename = curFilename;
                }
                // Determine the upper-left corner position of the pose within the source image
                sourceX = 0;
                sourceY = RPG::battlerAnimations[rAnim2Index]->poses[poseIndex]->battleCharsetId * poseHeight;
                // Draw the source pose into the Actor's pose
                rActorPtr->poses[poseIndex]->draw( 0, 0,                                // Coordinates in destination Image
                                   sourceImagePtr,                                      // Source Image pointer
                                   sourceX, sourceY,                                    // Coordinates in source Image
                                   poseWidth, poseHeight,                               // Dimensions in destination Image
                                   0 );                                                 // Transparency color
            }
        }
    }
    // Destroy the source Image
    RPG::Image::destroy( sourceImagePtr );
}

//! Update the mirror poses of a given Actor
/*!
    updateMirrorPoses takes the regular pose Images of an Actor and copies a mirrored version of
    them into the mirror pose Images of the same Actor. It is intended to be called after all of the
    Actor's regular pose Images have been modified in accordance with wardrobe data.

    \param rActorPtr: (RPG::Actor *) Pointer to the actor to draw upon
*/
void updateMirrorPoses( RPG::Actor * rActorPtr )
{
    static int poseIndex;                   // Index of the current pose
    static int x, y;                        // Coordinate variables
    static int width, height;

    // Loop through all pose images of the actor
    for( poseIndex = 0; poseIndex < NUM_ACTOR_POSES; poseIndex++ )
    {
        // Get width and height of pose image
        width = rActorPtr->poses[poseIndex]->width;
        height = rActorPtr->poses[poseIndex]->height;
        for( x = 0; x < width; x++ )
        {
            for( y = 0; y < height; y++ )
            {
                // Copy pixel from regular pose to horizontally symmetric pixel in mirrored pose
                rActorPtr->posesMirrored[poseIndex]->pixel( width - x, y ) = rActorPtr->poses[poseIndex]->pixel( x, y );
            }
        }
    }
}

//! Update the battle images of the actors currently in the party
/*!
    updateActors creates new battle images for the actors currently in the party based on the items
    the actors have equipped and the wardrobe associations defined in the plugin. This method is
    called automatically at the start of every battle, and can also be triggered manually by the
    dynwardrobe_update comment command.
*/
void updateActors()
{
    static int i, j;                    // Index variables
    static int itemIndex;               // RM2K3 database index of the item in the relevant equipment slot
    static bool baseDrawn;              // Whether the actor's base Animation2 has been drawn in yet
    static RPG::Actor *curActorPtr;     // Pointer to the current actor being updated
    static std::set<WardrobeAssociator>::iterator wardrobeIter;        // Iterator for traversing the set of WardrobeAssociators

    // Loop through all actors currently in the party
    for( i = 0; i < NUM_HEROES; i++ )
    {
        curActorPtr = RPG::Actor::partyMember( i );
        // Check for non-existent actor
        if( NULL != curActorPtr )
        {   // Actor exists
            baseDrawn = false;
            // Clear the actor's pose images
            for( j = 0; j < NUM_ACTOR_POSES; j++ )
            {
                curActorPtr->poses[j]->clear();
            }
            // Loop through all wardrobe associations
            for( wardrobeIter = wardrobe.begin(); wardrobeIter != wardrobe.end(); ++wardrobeIter )
            {
                // Make sure actor's base Animation2 gets drawn at the proper time
                if( !baseDrawn && 0 <= wardrobeIter->GetDrawingLayer() )
                {   // Base Animation2 has not yet been drawn and the current drawing layer is at least 0
                    // Draw base Animation2
                    drawAnim2( curActorPtr->getBattleGraphicId(), curActorPtr );
                    baseDrawn = true;
                }
                // Check whether the current WardrobeAssociator involves the current actor's base Animation2
                if( wardrobeIter->CountActorAnim2Indices( curActorPtr->getBattleGraphicId() ) )
                {   // Current actor's base Animation2 is involved in the WardrobeAssociator
                    // Find out which item the current actor has equipped in the current
                    // WardrobeAssociator's equipment slot
                    if( ACCESSORY_SLOT == wardrobeIter->GetEquipSlot() )
                        itemIndex = curActorPtr->accessoryId;
                    if( ARMOR_SLOT == wardrobeIter->GetEquipSlot() )
                        itemIndex = curActorPtr->armorId;
                    if( HELMET_SLOT == wardrobeIter->GetEquipSlot() )
                        itemIndex = curActorPtr->helmetId;
                    if( WEAPON_SLOT == wardrobeIter->GetEquipSlot() )
                        itemIndex = curActorPtr->weaponId;
                    if( SHIELD_SLOT == wardrobeIter->GetEquipSlot() )
                        itemIndex = curActorPtr->shieldId;
                    // Check whether the current WardrobeAssociator involves the item
                    if( wardrobeIter->CountItemIndices( itemIndex ) )
                    {   // Item is involved in the WardrobeAssociator
                        drawAnim2( wardrobeIter->GetItemAnim2Index(), curActorPtr );
                    }
                }
            }
            // Final check to make certain the actor's base Animation2 gets drawn in
            if( !baseDrawn )
            {   // Base Animation2 has not yet been drawn
                // Draw base Animation2
                drawAnim2( curActorPtr->getBattleGraphicId(), curActorPtr );
                baseDrawn = true;
            }
            // Update mirror poses
            updateMirrorPoses( curActorPtr );
        }
    }
}

//! Translate a string into a set of ints
/*!
    stringToIntSet takes a string, which is assumed to be data from DynWardrobe.csv representing a
    set of integer values, and returns a std::set<int> containing those integer values.

    \param rDataString : (std::string) String of input data
    \return (std::set<int>) The set of integer values represented by the input string
*/
std::set<int> stringToIntSet( std::string rDataString )
{
    static std::string dataString;      // Working string for the data
    static std::set<int> dataSet;       // Set of integer values to return
    static unsigned int tokenPos;       // End position of the current token in the data string

    // Clear the working data set
    dataSet.clear();
    // Copy the input string to a working string
    dataString = rDataString;
    // Strip the "" from the ends of the string if they're there
    if( '"' == dataString[0] )
    {   // " detected
        dataString = dataString.substr( 1, dataString.length() - 2 );
    }
    // Loop through all tokens in the data string
    for( tokenPos = dataString.find_first_of( ';' ); std::string::npos != tokenPos; tokenPos = dataString.find_first_of( ';' ) )
    {
        // Add the number of the current token to the data set
        dataSet.insert( std::atoi( ( dataString.substr( 0, tokenPos ) ).c_str() ) );
        // Remove the current token from the data string
        dataString.erase( 0, tokenPos + 1 );
    }
    // Add the final token (which does not have an associated ;) to the data set
    dataSet.insert( std::atoi( dataString.c_str() ) );
    // Return the data set
    return dataSet;
}

//! Translate a string into an equipment slot index
/*!
    stringToEquipSlot takes a string, which is assumed to be data from DynWardrobe.csv representing
    an equipment slot, and returns an EquipSlot for the appropriate equipment slot.

    \param rDataString : ( std::string ) String of input data
    \return (EquipSlot) The equipment slot index indicated by the data string
*/
EquipSlot stringToEquipSlot( std::string rDataString )
{
    // Determine which equipment slot is indicated by the data string
    if( 0 == rDataString.compare( "ACCESSORY" ) )
    {
        return ACCESSORY_SLOT;
    }
    if( 0 == rDataString.compare( "ARMOR" ) )
    {
        return ARMOR_SLOT;
    }
    if( 0 == rDataString.compare( "HELMET" ) )
    {
        return HELMET_SLOT;
    }
    if( 0 == rDataString.compare( "WEAPON" ) )
    {
        return WEAPON_SLOT;
    }
    if( 0 == rDataString.compare( "SHIELD" ) )
    {
        return SHIELD_SLOT;
    }
    // No equipment slot recognized; return invalid
    return INVALID_SLOT;
}

//! Called after the RPG objects were initialized
/*!
    onInitFinished is called when the objects of the RPG class are finished initializing. In this
    plugin this method is used to initialize variables that are dependent on RPG objects.

    \param pluginName : ( char * ) Name of the plugin
*/
void onInitFinished()
{
    #ifdef DYNWARDROBE_DEBUG
    ofDebug = new std::ofstream();
    ofDebug->open( "DynPlugins\\DynWardrobeDebug.txt" );
    #endif // DYNWARDROBE_DEBUG

    std::ifstream sourceFile( "DynPlugins\\DynWardrobe.csv" );      // File stream used to get source data
    CSVIterator lineTokens( sourceFile );       // Iterator for traversing source data
    WardrobeAssociator newAssociator;           // Temporary WardrobeAssociator
    std::string curToken;                       // Temporary token string
    std::set<int> newSet;                       // Temporary int set (used for several of WardrobeAssociator's attributes

    // Initialize variables
    inBattle = false;
    // Skip first line (headers) of source data
    ++lineTokens;
    // Get wardrobe association data
    for( ; lineTokens != CSVIterator(); ++lineTokens )
    {
        newAssociator.clear();
        curToken = (*lineTokens)[DRAWING_LAYER_CSV_INDEX];
        newAssociator.SetDrawingLayer( std::atoi( curToken.c_str() ) );
        curToken = (*lineTokens)[EQUIP_SLOT_CSV_INDEX];
        newAssociator.SetEquipSlot( stringToEquipSlot( curToken ) );
        curToken = (*lineTokens)[ITEM_ANIM2_CSV_INDEX];
        newAssociator.SetItemAnim2Index( std::atoi( curToken.c_str() ) );
        curToken = (*lineTokens)[ACTOR_ANIM2_CSV_INDEX];
        newAssociator.SetActorAnim2Indices( stringToIntSet( curToken ) );
        curToken = (*lineTokens)[ITEM_CSV_INDEX];
        newAssociator.SetItemIndices( stringToIntSet( curToken ) );
        if( newAssociator.IsValid() )
        {   // New associator is valid
            // Add the new associator to the wardrobe
            wardrobe.insert( newAssociator );
        }
    }
}

//! Called every frame
/*!
    onFrame is called on every frame of the game loop. In this plugin this method is used to
    detect and react to transitions between battles and other scenes.

    \param scene : ( RPG::Scene ) Enumerator for what type of scene the game is currently in
*/
void onFrame( RPG::Scene scene )
{
    if( inBattle )
    {   // Game was in a battle scene at last check
        if( RPG::SCENE_BATTLE != scene )
        {   // Current scene is not a battle; battle just ended!
            inBattle = false;
        }
    }
    else if( RPG::SCENE_BATTLE == scene )
    {   // Game was not in a battle scene at last check and current scene is a battle; battle just started!
        inBattle = true;
        updateActors();
    }
}

//! Respond to potential comment commands
/*!
    onComment is called when the game runs across a comment line in its event scripting. Comment
    commands are how the user communicates with the plugin, so this function checks for comment
    commands and responds accordingly.

    \param text (const char*) The comment's content as simple text
    \param parsedData (const RPG::ParsedCommentData*) The already parsed data
    \param nextScriptLine (RPG::EventScriptLine*) The next event script line after the comment
    \param scriptData (RPG::EventScriptData*) Pointer to the RPG::EventScriptData object of the current event script
    \param eventId (int) The ID of the current event (negative for common events, zero for battle events)
    \param pageId (int) The ID of the current event page (zero for common and battle events - sorry, no battle event page ID yet)
    \param lineId (int) The zero-based line number
    \param nextLineId (int*) Pointer to the next executed line number (-1 for default)
    \return (bool) false will prevent other plugins from receiving this notification, use true otherwise
*/
bool onComment( const char* text,
                const RPG::ParsedCommentData* 	parsedData,
                RPG::EventScriptLine* 	nextScriptLine,
                RPG::EventScriptData* 	scriptData,
                int 	eventId,
                int 	pageId,
                int 	lineId,
                int* 	nextLineId )
{
    static std::string cmd;

    // Check for and respond to plugin comment commands
    cmd = parsedData->command;
    if( 0 == cmd.compare( "dynwardrobe_update" ) )
    {   // Update the actors' battle images
        updateActors();
        return false;
    }
    // No comment commands for this plugin detected; pass notification on for other plugins
    return true;
}

#ifdef DYNWARDROBE_DEBUG
//! Clean up after use
/*!
    onExit is called when the game closes. In this plugin this is used to perform any needed
    cleanup.
*/
void onExit()
{
    ofDebug->close();
}
#endif // DYNWARDROBE_DEBUG
